//
// $Id: ipfw-graph.c,v 1.4 2001/10/08 00:16:29 mavetju Exp $
//

#include <gtk/gtk.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#define bool unsigned char

static GdkPixmap *drawing_pixmap = NULL;
static GdkPixmap *bytes_pixmap = NULL;
static GtkWidget *drawing_area;
static GtkWidget *bytes_area;

GdkGC *gc_red,*gc_yellow,*gc_green,*gc_cyan,*gc_blue,*gc_black,*gc_white;

typedef struct data_type	DATA_TYPE;

struct data_type {
    long new;
    long old;
    GdkGC *gc;
    long *table;
    char *rule;
    char *number;
};

DATA_TYPE data_bytes;		// data for the bytes received
DATA_TYPE *data=NULL;		// data for the ipfw rules
int	data_size;		// number of elements of data

int	xsize;			// xsize of the output
int	ysize_bytes,ysize_drawing;	// ysize of the bytes/piles output
long	*maxyvalues;		// maxyvalues per column in the output
long	maxyvalue=-1;		// biggest of the maxyvalues


//
// Get a new colour for the colourmap.
// Usage: GetPen(NewColour(red,green,blue))
//
// Credits: found on the Internet somewhere
//
GdkGC *GetPen(GdkColor *c) {
    GdkGC *gc;

    gc=gdk_gc_new(drawing_pixmap);
    gdk_gc_set_foreground(gc,c);
    return gc;
}

GdkColor *NewColour(long red,long green,long blue) {
    GdkColor *c=(GdkColor *)g_malloc(sizeof(GdkColor));

    c->red=red;
    c->green=green;
    c->blue=blue;

    gdk_color_alloc(gdk_colormap_get_system(),c);
    return c;
}


//
// Called after a resize of the window
//
/* Create a new backing pixmap of the appropriate size */
static gint configure_bytes_event(GtkWidget *widget,GdkEventConfigure *event) {
    if (bytes_pixmap)
	gdk_pixmap_unref(bytes_pixmap);

    ysize_bytes=bytes_area->allocation.height;

    bytes_pixmap = gdk_pixmap_new(widget->window,
			    widget->allocation.width,
			    widget->allocation.height,
			    -1);
    gdk_draw_rectangle (bytes_pixmap,
			widget->style->white_gc,
			TRUE,
			0, 0,
			widget->allocation.width,
			widget->allocation.height);
    return TRUE;
}

void recreate_table(DATA_TYPE *data,int old,int new) {
    long *	old_table;
    old_table=(long *)calloc(old,sizeof(long));

    memmove(old_table,data->table,old*sizeof(long));
    free(data->table);
    data->table=calloc(new,sizeof(long));

    if (old<new)
	memmove(data->table+(new-old),old_table,old*sizeof(long));
    else
	memmove(data->table,old_table+(old-new),new*sizeof(long));
    free(old_table);
}

static gint configure_drawing_event(GtkWidget *widget,GdkEventConfigure *event) {
    if (drawing_pixmap)
	gdk_pixmap_unref(drawing_pixmap);

    if (maxyvalue>=0) {
	int	old_xsize;
	long *	old_table;
	int	i;

	old_xsize=xsize;
	old_table=(long *)calloc(xsize,sizeof(long));

	xsize=drawing_area->allocation.width;
	ysize_drawing=drawing_area->allocation.height;

	memmove(old_table,maxyvalues,old_xsize*sizeof(long));
	free(maxyvalues);
	maxyvalues=(long *)calloc(xsize,sizeof(long));
	for (i=0;i<xsize;i++) maxyvalues[i]=1;
	if (old_xsize<xsize)
	    memmove(maxyvalues+(xsize-old_xsize),old_table,old_xsize*sizeof(long));
	else
	    memmove(maxyvalues,old_table+(old_xsize-xsize),xsize*sizeof(long));
	free(old_table);

	recreate_table(&data_bytes,old_xsize,xsize);
	for (i=0;i<data_size;i++)
	    recreate_table(&data[i],old_xsize,xsize);
    }

    drawing_pixmap = gdk_pixmap_new(widget->window,
			    widget->allocation.width,
			    widget->allocation.height,
			    -1);
    gdk_draw_rectangle (drawing_pixmap,
			widget->style->white_gc,
			TRUE,
			0, 0,
			widget->allocation.width,
			widget->allocation.height);
    return TRUE;
}

//
// Redraw the screen from the backing pixmap (???)
//
static gint expose_bytes_event (GtkWidget *widget, GdkEventExpose *event) {
    gdk_draw_pixmap(widget->window,
		    widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
		    bytes_pixmap,
		    event->area.x, event->area.y,
		    event->area.x, event->area.y,
		    event->area.width, event->area.height);
    return FALSE;
}

static gint expose_drawing_event (GtkWidget *widget, GdkEventExpose *event) {
    gdk_draw_pixmap(widget->window,
		    widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
		    drawing_pixmap,
		    event->area.x, event->area.y,
		    event->area.x, event->area.y,
		    event->area.width, event->area.height);
    return FALSE;
}

//
// Button press handling
//
static void print_button_press (guint32 deviceid) {
    GList *tmp_list;

    /* gdk_input_list_devices returns an internal list, so we shouldn't
    free it afterwards */
    tmp_list = gdk_input_list_devices();

    while (tmp_list) {
	GdkDeviceInfo *info = (GdkDeviceInfo *)tmp_list->data;

	if (info->deviceid == deviceid) {
	    g_print("Button press on device '%s'\n", info->name);
	    return;
	}

	tmp_list = tmp_list->next;
    }
}

static gint button_press_event (GtkWidget *widget, GdkEventButton *event) {
    print_button_press (event->deviceid);
    return TRUE;
}

//
// Input dialog (???)
//
void input_dialog_destroy (GtkWidget *w, gpointer data) {
    *((GtkWidget **)data) = NULL;
}

void create_input_dialog(void) {
    static GtkWidget *inputd = NULL;

    if (!inputd) {
	inputd = gtk_input_dialog_new();

	gtk_signal_connect (GTK_OBJECT(inputd), "destroy",
			    (GtkSignalFunc)input_dialog_destroy, &inputd);
	gtk_signal_connect_object (GTK_OBJECT(GTK_INPUT_DIALOG(inputd)->close_button),
				"clicked",
				(GtkSignalFunc)gtk_widget_hide,
				GTK_OBJECT(inputd));
	gtk_widget_hide ( GTK_INPUT_DIALOG(inputd)->save_button);

	gtk_widget_show (inputd);
    } else {
	if (!GTK_WIDGET_MAPPED(inputd))
	    gtk_widget_show(inputd);
	else
	    gdk_window_raise(inputd->window);
    }
}

//
// Quit button
//
void quit(void) {
    gtk_exit(0);
}


//
// Add 'value' to the 'table' in 'data'
// If updatemax==TRUE, then update the global variable maxyvalue also
//
void update_data(long value,DATA_TYPE *data,bool updatemax) {
    if (maxyvalue==-1)
	data->new=value;

    data->old=data->new;
    data->new=value;

    memmove(data->table,data->table+1,sizeof(int)*(xsize-1));

    data->table[xsize-1]=data->new-data->old;

    if (updatemax) {
	int i;

	memmove(maxyvalues,maxyvalues+1,(xsize-1)*sizeof(long));
	maxyvalues[xsize-1]=data->table[xsize-1]==0?1:data->table[xsize-1];

	maxyvalue=1;
	for (i=0;i<xsize;i++)
	    if (maxyvalue<maxyvalues[i])
		maxyvalue=maxyvalues[i];
    }
}

//
// Find a certain rule
//
DATA_TYPE *find_data(char *number,char *rule) {
    int i;

    for (i=0;i<data_size;i++)
	if (strcmp(data[i].number,number)==0 &&
	    strcmp(data[i].rule,rule)==0)
	    return &data[i];
    return NULL;
}

//
// Get the data from the input source and store in into the data-variable
//
#define STRINGSIZE	1000
void getdata(void) {
    FILE *	fin;
    int		packets,bytes;
    int		totalpackets,totalbytes;
    int		chars;
    char	s[STRINGSIZE];
    char	number[STRINGSIZE];
    char *	rule;
    DATA_TYPE *	thisdata;

    totalpackets=0;
    totalbytes=0;

    if ((fin=popen("./ipfw-al","r"))!=NULL) {
	while (!feof(fin)) {
	    if (fgets(s,STRINGSIZE,fin)==NULL) break;
	    if (sscanf(s,"%s %d %d %n",number,&packets,&bytes,&chars)==3) {
		s[chars-1]=0;
		rule=s+chars;
		rule[strlen(rule)-1]=0;

		if ((thisdata=find_data(number,rule))!=NULL)
		    update_data(bytes,thisdata,FALSE);
		else
		    printf("Didn't find: '%s' '%s'\n",number,rule);
		totalbytes+=bytes;
		totalpackets+=packets;
	    }
	}
	pclose(fin);
    }

    update_data(totalbytes,&data_bytes,TRUE);
}


//
// Draw the graphs
//
long *piles;
void draw_init(void) {
    piles=(long *)calloc(xsize,sizeof(long));
}

// draw the coloured piles on top of each other
void draw_pile(DATA_TYPE data) {
    int x,y;

    for (x=0;x<xsize;x++) {
	y=piles[x];
	gdk_draw_line(drawing_pixmap,data.gc,
		x,ysize_drawing-(ysize_drawing*(y              ))/maxyvalues[x],
		x,ysize_drawing-(ysize_drawing*(y+data.table[x]))/maxyvalues[x]);
	piles[x]+=data.table[x];
    }

}

// draw a realtime graph
void draw_line(DATA_TYPE data) {
    int x;

    for (x=1;x<xsize;x++)
	gdk_draw_line(bytes_pixmap,data.gc,
		x-1,ysize_bytes-(ysize_bytes*data.table[x-1])/maxyvalue,
		x  ,ysize_bytes-(ysize_bytes*data.table[x  ])/maxyvalue);
}

void draw_end(void) {
    free(piles);
}

// draw the tables
void draw_tables(void) {
    int i;

    draw_init();
    for (i=0;i<data_size;i++) {
	draw_pile(data[i]);
    }
    draw_end();
    draw_line(data_bytes);
}

//
// periodic called function to redraw the screen
//
gint update(void) {
    GdkGC *gc;
    GdkRectangle update_rect;
    int y;

    //
    // bytes part
    //
    update_rect.x=0;
    update_rect.y=0;
    update_rect.width=bytes_area->allocation.width;
    update_rect.height=bytes_area->allocation.height;

    gc = bytes_area->style->bg_gc[GTK_WIDGET_STATE (bytes_area)];

    gdk_draw_rectangle(bytes_pixmap,gc,TRUE,
			update_rect.x, update_rect.y,
			update_rect.width, update_rect.height);

    gc = bytes_area->style->black_gc;
    for (y=0;y<10;y++) {
	gdk_draw_line(bytes_pixmap,gc,
		1    ,y*ysize_bytes/10,
		xsize,y*ysize_bytes/10);
    }

    //
    // drawing part
    //
    update_rect.x=0;
    update_rect.y=0;
    update_rect.width=drawing_area->allocation.width;
    update_rect.height=drawing_area->allocation.height;

    gc = drawing_area->style->bg_gc[GTK_WIDGET_STATE (drawing_area)];

    gdk_draw_rectangle(drawing_pixmap,gc,TRUE,
			update_rect.x, update_rect.y,
			update_rect.width, update_rect.height);

    //
    // graphs
    //
    getdata();
    draw_tables();
    gtk_widget_queue_draw(drawing_area);
    gtk_widget_queue_draw(bytes_area);

    return 1;
}

//
// Read the config from ipfw
//
void get_config(void) {
    FILE *fin;
    char s[1000];
    char *p;

    if ((fin=popen("./ipfw-al defs","r"))==NULL) {
	perror("ipfw-al defs");
	exit(0);
    }

    while (!feof(fin)) {
	if (fgets(s,sizeof(s),fin)==NULL)
	    break;
	if (s[0]==0) continue;
	if (s[0]=='#') continue;

	s[strlen(s)-1]=0;

	data=(DATA_TYPE *)realloc(data,(data_size+1)*sizeof(DATA_TYPE));

	data[data_size].new=0;
	data[data_size].old=0;
	data[data_size].gc=NULL;
	data[data_size].table=NULL;

	if ((p=strchr(s,' '))==NULL) {
	    perror(s);
	    exit(0);
	}
	*p=0;
	p++;
	data[data_size].number=strdup(s);
	data[data_size].rule=strdup(p);

	data_size++;
    }
    pclose(fin);
}

int main (int argc, char *argv[]) {
    GtkWidget *window;
    GtkWidget *vbox;

    GtkWidget *button;
    GtkStyle *style;

    gtk_init(&argc,&argv);

    window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_widget_set_name(window,"Test Input");

    vbox=gtk_vbox_new(FALSE,0);
    gtk_container_add(GTK_CONTAINER(window),vbox);
    gtk_widget_show(vbox);

    gtk_signal_connect(GTK_OBJECT(window),"destroy",GTK_SIGNAL_FUNC(quit),NULL);

    /* Create the drawing area */
    drawing_area=gtk_drawing_area_new();
    gtk_drawing_area_size(GTK_DRAWING_AREA(drawing_area),200,300);
    gtk_box_pack_start(GTK_BOX(vbox),drawing_area,TRUE,TRUE,0);
    gtk_widget_show(drawing_area);

    bytes_area=gtk_drawing_area_new();
    gtk_drawing_area_size(GTK_DRAWING_AREA(bytes_area),200,50);
    gtk_box_pack_start(GTK_BOX(vbox),bytes_area,TRUE,TRUE,0);
    gtk_widget_show(bytes_area);

    /* Signals used to handle backing pixmap */
    gtk_signal_connect(GTK_OBJECT(drawing_area),"expose_event",
			(GtkSignalFunc)expose_drawing_event,NULL);
    gtk_signal_connect(GTK_OBJECT(bytes_area),"expose_event",
			(GtkSignalFunc)expose_bytes_event,NULL);
    gtk_signal_connect(GTK_OBJECT(drawing_area),"configure_event",
			(GtkSignalFunc)configure_drawing_event,NULL);
    gtk_signal_connect(GTK_OBJECT(bytes_area),"configure_event",
			(GtkSignalFunc)configure_bytes_event,NULL);

    /* Event signals */
    gtk_signal_connect(GTK_OBJECT(drawing_area),"button_press_event",
			(GtkSignalFunc)button_press_event,NULL);
    gtk_signal_connect(GTK_OBJECT(bytes_area),"button_press_event",
			(GtkSignalFunc)button_press_event,NULL);

    gtk_widget_set_events(drawing_area,   GDK_EXPOSURE_MASK
					| GDK_LEAVE_NOTIFY_MASK
					| GDK_BUTTON_PRESS_MASK
					| GDK_POINTER_MOTION_MASK
					| GDK_POINTER_MOTION_HINT_MASK);

    /* The following call enables tracking and processing of extension
	events for the drawing area */
    gtk_widget_set_extension_events(drawing_area,GDK_EXTENSION_EVENTS_CURSOR);

    /* .. And some buttons */
    button=gtk_button_new_with_label("Quit");
    gtk_box_pack_start(GTK_BOX(vbox),button,FALSE,FALSE,0);

    gtk_signal_connect_object(GTK_OBJECT(button),"clicked",
				GTK_SIGNAL_FUNC(gtk_widget_destroy),
				GTK_OBJECT(window));
    gtk_widget_show(button);
    gtk_widget_show(window);

    style=gtk_widget_get_default_style();
    gc_red=GetPen(NewColour(0xffff,0,0));
    gc_yellow=GetPen(NewColour(0xffff,0xffff,0));
    gc_green=GetPen(NewColour(0,0xffff,0));
    gc_cyan=GetPen(NewColour(0,0xffff,0xffff));
    gc_blue=GetPen(NewColour(0,0,0xffff));
    gc_black=GetPen(NewColour(0,0,0));
    gc_white=GetPen(NewColour(0xffff,0xffff,0xffff));

    xsize=drawing_area->allocation.width;
    ysize_drawing=drawing_area->allocation.height;
    ysize_bytes=bytes_area->allocation.height;

    //
    // init
    //
    {
	int i;

	maxyvalues=(long *)calloc(xsize,sizeof(long));
	for (i=0;i<xsize;i++) maxyvalues[i]=1;

	data=NULL;
	data_bytes.new=0;
	data_bytes.old=0;
	data_bytes.rule=NULL;
	data_bytes.number=NULL;

	data_size=0;
	get_config();


	data_bytes.gc=gc_white;
	data_bytes.table=calloc(xsize,sizeof(long));
	for (i=0;i<data_size;i++) {
	    data[i].gc=GetPen(NewColour(random()%0xffff,
				       random()%0xffff,
				       random()%0xffff));
	    data[i].table=calloc(xsize,sizeof(long));
	}
    }

    gtk_timeout_add(5000,(GtkFunction)update,NULL);

    gtk_main();

    return 0;
}
